@@ -8,10 +8,11 @@ from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import unicode_literals
 
-from datetime import datetime
 import time
-import zipfile
 import tarfile
+import typing
+import zipfile
+from datetime import datetime
 
 import six
 
@@ -21,21 +22,29 @@ from .time import datetime_to_epoch
 from .errors import NoSysPath, MissingInfoNamespace
 from .walk import Walker
 
+if False:  # typing.TYPE_CHECKING
+    from typing import BinaryIO, Optional, Text, Tuple, Type, Union
+    from .base import FS
+    ZipTime = Tuple[int, int, int, int, int, int]
 
-def write_zip(src_fs,
-              file,
-              compression=zipfile.ZIP_DEFLATED,
-              encoding="utf-8",
-              walker=None):
+
+def write_zip(src_fs,                            # type: FS
+              file,                              # type: Union[Text, BinaryIO]
+              compression=zipfile.ZIP_DEFLATED,  # type: int
+              encoding="utf-8",                  # type: Text
+              walker=None                        # type: Optional[Walker]
+              ):
+    # type: (...) -> None
     """Write the contents of a filesystem to a zip file.
 
     Arguments:
+        src_fs (~fs.base.FS): The source filesystem to compress.
         file (str or io.IOBase): Destination file, may be a file name
             or an open file object.
-        compression (str, optional): Compression to use (one of the constants
+        compression (int): Compression to use (one of the constants
             defined in the `zipfile` module in the stdlib). Defaults
             to `zipfile.ZIP_DEFLATED`.
-        encoding (str, optional):
+        encoding (str):
              The encoding to use for filenames. The default is ``"utf-8"``,
              use ``"CP437"`` if compatibility with WinZip is desired.
         walker (~fs.walk.Walker, optional): A `Walker` instance, or `None`
@@ -66,7 +75,7 @@ def write_zip(src_fs,
                 # zip time directory from the stat structure
                 st_mtime = info.get('stat', 'st_mtime', None)
                 _mtime = time.localtime(st_mtime)
-                zip_time = _mtime[0:6]
+                zip_time = _mtime[0:6]  # type: ZipTime
             else:
                 # Otherwise, use the modified time from details
                 # namespace.
@@ -75,7 +84,10 @@ def write_zip(src_fs,
                     mt.year, mt.month, mt.day,
                     mt.hour, mt.minute, mt.second
                 )
-            zip_info = zipfile.ZipInfo(zip_name, zip_time)
+
+            # NOTE(@althonos): typeshed's `zipfile.py` on declares
+            #     ZipInfo.__init__ for Python < 3 ?!
+            zip_info = zipfile.ZipInfo(zip_name, zip_time)  # type: ignore
 
             try:
                 zip_info.external_attr = info.permissions.mode << 16
@@ -102,23 +114,25 @@ def write_zip(src_fs,
                     _zip.write(sys_path, zip_name)
 
 
-def write_tar(src_fs,
-              file,
-              compression=None,
-              encoding="utf-8",
-              walker=None):
+def write_tar(src_fs,            # type: FS
+              file,              # type: Union[Text, BinaryIO]
+              compression=None,  # type: Optional[Text]
+              encoding="utf-8",  # type: Text
+              walker=None        # type: Optional[Walker]
+              ):
+    # type: (...) -> None
     """Write the contents of a filesystem to a tar file.
 
     Arguments:
-        file (str or io.IOBase): Destination file, may be a file name or
-            an open file object.
+        file (str or io.IOBase): Destination file, may be a file
+            name or an open file object.
         compression (str, optional): Compression to use, or `None`
             for a plain Tar archive without compression.
-        encoding(str, optional): The encoding to use for filenames. The
+        encoding(str): The encoding to use for filenames. The
             default is ``"utf-8"``.
-        walker (~fs.walk.Walker, optional): A `Walker` instance, or `None`
-            to use default walker. You can use this to specify which files
-            you want to compress.
+        walker (~fs.walk.Walker, optional): A `Walker` instance, or
+            `None` to use default walker. You can use this to specify
+            which files you want to compress.
 
     """
     type_map = {
@@ -140,10 +154,10 @@ def write_tar(src_fs,
     ]
 
     mode = 'w:{}'.format(compression or '')
-    if hasattr(file, 'write'):
-        _tar = tarfile.open(fileobj=file, mode=mode)
-    else:
+    if isinstance(file, (six.text_type, six.binary_type)):
         _tar = tarfile.open(file, mode=mode)
+    else:
+        _tar = tarfile.open(fileobj=file, mode=mode)
 
     current_time = time.time()
     walker = walker or Walker()
