@@ -3,6 +3,9 @@
 
 from __future__ import unicode_literals
 
+import copy
+import typing
+
 import six
 
 from . import errors
@@ -13,9 +16,29 @@ from .move import move_file
 from .path import abspath, normpath
 from .error_tools import unwrap_errors
 
+if False:  # typing.TYPE_CHECKING
+    from datetime import datetime
+    from threading import RLock
+    from typing import (
+        Any, AnyStr, BinaryIO, Callable, Collection, Dict,
+        Iterator, Iterable, IO, List, Mapping, Optional,
+        Text, TextIO, Tuple, Union)
+    from .enums import ResourceType
+    from .info import RawInfo
+    from .permissions import Permissions
+    from .subfs import SubFS
+    from .walk import BoundWalker
+
+    _T = typing.TypeVar('_T', bound='FS')
+    _OpendirFactory = Callable[[_T, Text], SubFS[_T]]
+
+
+_F = typing.TypeVar('_F', bound='FS', covariant=True)
+_W = typing.TypeVar('_W', bound='WrapFS[FS]')
+
 
 @six.python_2_unicode_compatible
-class WrapFS(FS):
+class WrapFS(FS, typing.Generic[_F]):
     """A proxy for a filesystem object.
 
     This class exposes an filesystem interface, where the data is
@@ -24,26 +47,29 @@ class WrapFS(FS):
 
     """
 
-    wrap_name = None
+    wrap_name = None    # type: Optional[Text]
 
     def __init__(self, wrap_fs):
+        # type: (_F) -> None
         self._wrap_fs = wrap_fs
         super(WrapFS, self).__init__()
 
     def __repr__(self):
+        # type: () -> Text
         return "{}({!r})".format(
             self.__class__.__name__,
             self._wrap_fs
         )
 
     def __str__(self):
+        # type: () -> Text
         wraps = []
-        _fs = self
+        _fs = self   # type: Union[FS, WrapFS[FS]]
         while hasattr(_fs, '_wrap_fs'):
             wrap_name = getattr(_fs, 'wrap_name', None)
             if wrap_name is not None:
                 wraps.append(wrap_name)
-            _fs = _fs._wrap_fs
+            _fs = _fs._wrap_fs  # type: ignore
         if wraps:
             _str = "{}({})".format(_fs, ', '.join(wraps[::-1]))
         else:
@@ -51,6 +77,7 @@ class WrapFS(FS):
         return _str
 
     def delegate_path(self, path):
+        # type: (Text) -> Tuple[_F, Text]
         """Encode a path for proxied filesystem.
 
         Arguments:
@@ -63,6 +90,7 @@ class WrapFS(FS):
         return self._wrap_fs, path
 
     def delegate_fs(self):
+        # type: () -> _F
         """Get the proxied filesystem.
 
         This method should return a filesystem for methods not
@@ -72,13 +100,20 @@ class WrapFS(FS):
         return self._wrap_fs
 
     def appendbytes(self, path, data):
+        # type: (Text, bytes) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             return _fs.appendbytes(_path, data)
 
-    def appendtext(self, path, text,
-                   encoding='utf-8', errors=None, newline=''):
+    def appendtext(self,
+                   path,                    # type: Text
+                   text,                    # type: Text
+                   encoding='utf-8',        # type: Text
+                   errors=None,             # type: Optional[Text]
+                   newline=''               # type: Text
+                   ):
+        # type: (...) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -89,17 +124,18 @@ class WrapFS(FS):
                                   newline=newline)
 
     def getinfo(self, path, namespaces=None):
+        # type: (Text, Optional[Collection[Text]]) -> Info
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             raw_info = _fs.getinfo(_path, namespaces=namespaces).raw
         if abspath(normpath(path)) == '/':
-            raw_info = raw_info.copy()
-            raw_info['basic']['name'] = ''
-        info = Info(raw_info)
-        return info
+            raw_info = dict(raw_info)
+            raw_info['basic']['name'] = ''  # type: ignore
+        return Info(raw_info)
 
     def listdir(self, path):
+        # type: (Text) -> List[Text]
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -107,11 +143,17 @@ class WrapFS(FS):
         return dir_list
 
     def lock(self):
+        # type: () -> RLock
         self.check()
         _fs = self.delegate_fs()
         return _fs.lock()
 
-    def makedir(self, path, permissions=None, recreate=False):
+    def makedir(self,
+                path,               # type: Text
+                permissions=None,   # type: Optional[Permissions]
+                recreate=False      # type: bool
+                ):
+        # type: (...) -> SubFS[FS]
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -122,6 +164,7 @@ class WrapFS(FS):
             )
 
     def move(self, src_path, dst_path, overwrite=False):
+        # type: (Text, Text, bool) -> None
         # A custom move permits a potentially optimized code path
         src_fs, _src_path = self.delegate_path(src_path)
         dst_fs, _dst_path = self.delegate_path(dst_path)
@@ -131,6 +174,7 @@ class WrapFS(FS):
             move_file(src_fs, _src_path, dst_fs, _dst_path)
 
     def openbin(self, path, mode='r', buffering=-1, **options):
+        # type: (Text, Text, int, **Any) -> BinaryIO
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -143,12 +187,14 @@ class WrapFS(FS):
         return bin_file
 
     def remove(self, path):
+        # type: (Text) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.remove(_path)
 
     def removedir(self, path):
+        # type: (Text) -> None
         self.check()
         _path = abspath(normpath(path))
         if _path == '/':
@@ -157,7 +203,12 @@ class WrapFS(FS):
         with unwrap_errors(path):
             _fs.removedir(_path)
 
-    def scandir(self, path, namespaces=None, page=None):
+    def scandir(self,
+                path,               # type: Text
+                namespaces=None,    # type: Optional[Collection[Text]]
+                page=None           # type: Optional[Tuple[int, int]]
+                ):
+        # type: (...) -> Iterator[Info]
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -167,23 +218,27 @@ class WrapFS(FS):
                 yield info
 
     def setinfo(self, path, info):
+        # type: (Text, RawInfo) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         return _fs.setinfo(_path, info)
 
     def settimes(self, path, accessed=None, modified=None):
+        # type: (Text, Optional[datetime], Optional[datetime]) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.settimes(_path, accessed=accessed, modified=modified)
 
     def touch(self, path):
+        # type: (Text) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.touch(_path)
 
     def copy(self, src_path, dst_path, overwrite=False):
+        # type: (Text, Text, bool) -> None
         src_fs, _src_path = self.delegate_path(src_path)
         dst_fs, _dst_path = self.delegate_path(dst_path)
         with unwrap_errors({_src_path: src_path, _dst_path: dst_path}):
@@ -192,12 +247,14 @@ class WrapFS(FS):
             copy_file(src_fs, _src_path, dst_fs, _dst_path)
 
     def create(self, path, wipe=False):
+        # type: (Text, bool) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
-            _fs.create(_path, wipe=wipe)
+            return _fs.create(_path, wipe=wipe)
 
     def desc(self, path):
+        # type: (Text) -> Text
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -205,6 +262,7 @@ class WrapFS(FS):
         return desc
 
     def exists(self, path):
+        # type: (Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -212,13 +270,15 @@ class WrapFS(FS):
         return exists
 
     def filterdir(self,
-                  path,
-                  exclude_dirs=False,
-                  exclude_files=False,
-                  files=None,
-                  dirs=None,
-                  namespaces=None,
-                  page=None):
+                  path,                 # type: Text
+                  files=None,           # type: Optional[Iterable[Text]]
+                  dirs=None,            # type: Optional[Iterable[Text]]
+                  exclude_dirs=None,    # type: Optional[Iterable[Text]]
+                  exclude_files=None,   # type: Optional[Iterable[Text]]
+                  namespaces=None,      # type: Optional[Collection[Text]]
+                  page=None             # type: Optional[Tuple[int, int]]
+                  ):
+        # type: (...) -> Iterator[Info]
         self.check()
         _fs, _path = self.delegate_path(path)
         iter_files = iter(_fs.filterdir(
@@ -235,13 +295,20 @@ class WrapFS(FS):
                 yield info
 
     def getbytes(self, path):
+        # type: (Text) -> bytes
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _bytes = _fs.getbytes(_path)
         return _bytes
 
-    def gettext(self, path, encoding=None, errors=None, newline=''):
+    def gettext(self,
+                path,           # type: Text
+                encoding=None,  # type: Optional[Text]
+                errors=None,    # type: Optional[Text]
+                newline=''      # type: Text
+                ):
+        # type: (...) -> Text
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -252,11 +319,13 @@ class WrapFS(FS):
         return _text
 
     def getmeta(self, namespace='standard'):
+        # type: (Text) -> Mapping[Text, object]
         self.check()
         meta = self.delegate_fs().getmeta(namespace=namespace)
         return meta
 
     def getsize(self, path):
+        # type: (Text) -> int
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -264,6 +333,7 @@ class WrapFS(FS):
         return size
 
     def getsyspath(self, path):
+        # type: (Text) -> Text
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -271,6 +341,7 @@ class WrapFS(FS):
         return sys_path
 
     def gettype(self, path):
+        # type: (Text) -> ResourceType
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -278,12 +349,14 @@ class WrapFS(FS):
         return _type
 
     def geturl(self, path, purpose='download'):
+        # type: (Text, Text) -> Text
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             return _fs.geturl(_path, purpose=purpose)
 
     def hassyspath(self, path):
+        # type: (Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -291,6 +364,7 @@ class WrapFS(FS):
         return has_sys_path
 
     def hasurl(self, path, purpose='download'):
+        # type: (Text, Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -298,6 +372,7 @@ class WrapFS(FS):
         return has_url
 
     def isdir(self, path):
+        # type: (Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -305,6 +380,7 @@ class WrapFS(FS):
         return _isdir
 
     def isfile(self, path):
+        # type: (Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -312,13 +388,19 @@ class WrapFS(FS):
         return _isfile
 
     def islink(self, path):
+        # type: (Text) -> bool
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _islink = _fs.islink(_path)
         return _islink
 
-    def makedirs(self, path, permissions=None, recreate=False):
+    def makedirs(self,
+                 path,              # type: Text
+                 permissions=None,  # type: Optional[Permissions]
+                 recreate=False     # type: bool
+                 ):
+        # type: (...) -> SubFS[FS]
         self.check()
         _fs, _path = self.delegate_path(path)
         return _fs.makedirs(
@@ -327,15 +409,18 @@ class WrapFS(FS):
             recreate=recreate
         )
 
+    # FIXME(@althonos): line_buffering is not a FS.open declared argument
     def open(self,
-             path,
-             mode='r',
-             buffering=-1,
-             encoding=None,
-             errors=None,
-             newline='',
-             line_buffering=False,
-             **options):
+             path,                      # type: Text
+             mode='r',                  # type: Text
+             buffering=-1,              # type: int
+             encoding=None,             # type: Optional[Text]
+             errors=None,               # type: Optional[Text]
+             newline='',                # type: Text
+             line_buffering=False,      # type: bool
+             **options                  # type: Any
+             ):
+        # type: (...) -> IO[AnyStr]
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -351,7 +436,11 @@ class WrapFS(FS):
             )
         return open_file
 
-    def opendir(self, path, factory=None):
+    def opendir(self,         # type: _W
+                path,         # type: Text
+                factory=None  # type: Optional[_OpendirFactory]
+                ):
+        # type: (...) -> SubFS[_W]
         from .subfs import SubFS
         factory = factory or SubFS
         if not self.getinfo(path).is_dir:
@@ -362,23 +451,27 @@ class WrapFS(FS):
             return factory(self, path)
 
     def setbytes(self, path, contents):
+        # type: (Text, bytes) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.setbytes(_path, contents)
 
     def setbinfile(self, path, file):
+        # type: (Text, BinaryIO) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.setbinfile(_path, file)
 
     def setfile(self,
-                path,
-                file,
-                encoding=None,
-                errors=None,
-                newline=''):
+                path,           # type: Text
+                file,           # type: IO[AnyStr]
+                encoding=None,  # type: Optional[Text]
+                errors=None,    # type: Optional[Text]
+                newline=''      # type: Text
+                ):
+        # type: (...) -> None
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
@@ -391,9 +484,15 @@ class WrapFS(FS):
             )
 
     def validatepath(self, path):
+        # type: (Text) -> Text
         self.check()
         _fs, _path = self.delegate_path(path)
         with unwrap_errors(path):
             _fs.validatepath(_path)
         path = abspath(normpath(path))
         return path
+
+    @property
+    def walk(self):
+        # type: () -> BoundWalker
+        return self._wrap_fs.walker_class.bind(self)
